@@ -91,18 +91,6 @@ struct origin {
 	struct list_head snapshots;
 };
 
-/*
- * Useful macro for running the store functions.  Use
- * store_int_fn if you want the return value.
- */
-#define store_fn(snap, fn, args...) \
-    if ((snap)->store. ## fn) \
-        (snap)->store. ## fn ( &(snap)->store , ## args )
-
-#define store_int_fn(snap, fn, args...) \
-    (((snap)->store. ## fn) ? \
-        ((snap)->store. ## fn ( &(snap)->store , ## args )) : 0)
-
 /*
  * Size of the hash table for origin volumes. If we make this
  * the size of the minors list then it should be nearly perfect
@@ -220,7 +208,7 @@ static int init_exception_table(struct exception_table *et, uint32_t size)
 	int i;
 
 	et->hash_mask = size - 1;
-	et->table = vmalloc(sizeof(struct list_head) * (size));
+	et->table = vcalloc(size, sizeof(struct list_head));
 	if (!et->table)
 		return -ENOMEM;
 
@@ -392,15 +380,23 @@ static int init_hash_tables(struct dm_snapshot *s)
 }
 
 /*
- * Construct a snapshot mapping: <origin_dev> <COW-dev> <p/n>
- * <chunk-size>
+ * Round a number up to the nearest 'size' boundary.  size must
+ * be a power of 2.
+ */
+static inline ulong round_up(ulong n, ulong size)
+{
+	size--;
+	return (n + size) & ~size;
+}
+
+/*
+ * Construct a snapshot mapping: <origin_dev> <COW-dev> <p/n> <chunk-size>
  */
 static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 			int argc, char **argv, void **context)
 {
 	struct dm_snapshot *s;
 	unsigned long chunk_size;
-	unsigned long extent_size = 0L;
 	int r = -EINVAL;
 	char *persistent;
 	char *origin_path;
@@ -409,7 +405,7 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 	int blocksize;
 
 	if (argc < 4) {
-		*context = "dm-snapshot: Not enough arguments";
+		*context = "dm-snapshot: requires exactly 4 arguments";
 		r = -EINVAL;
 		goto bad;
 	}
@@ -431,22 +427,6 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 		goto bad;
 	}
 
-	/* Get the extent size for persistent snapshots */
-	if ((*persistent & 0x5f) == 'P') {
-		if (argc < 5) {
-			*context = "No extent size specified";
-			r = -EINVAL;
-			goto bad;
-		}
-
-		extent_size = simple_strtoul(argv[4], &value, 10);
-		if (extent_size == 0 || value == NULL) {
-			*context = "Invalid extent size";
-			r = -EINVAL;
-			goto bad;
-		}
-	}
-
 	s = kmalloc(sizeof(*s), GFP_KERNEL);
 	if (s == NULL) {
 		*context = "Cannot allocate snapshot context private structure";
@@ -454,20 +434,27 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 		goto bad;
 	}
 
-	r = dm_table_get_device(t, origin_path, 0, 0, &s->origin);
+	r = dm_table_get_device(t, origin_path, 0, 0, FMODE_READ, &s->origin);
 	if (r) {
 		*context = "Cannot get origin device";
 		goto bad_free;
 	}
 
-	r = dm_table_get_device(t, cow_path, 0, 0, &s->cow);
+	r = dm_table_get_device(t, cow_path, 0, 0,
+				FMODE_READ | FMODE_WRITE, &s->cow);
 	if (r) {
 		dm_table_put_device(t, s->origin);
 		*context = "Cannot get COW device";
 		goto bad_free;
 	}
 
-	/* Validate the extent and chunk sizes against the device block size */
+	/*
+	 * Chunk size must be multiple of page size.  Silently
+	 * round up if it's not.
+	 */
+	chunk_size = round_up(chunk_size, PAGE_SIZE / SECTOR_SIZE);
+
+	/* Validate the chunk size against the device block size */
 	blocksize = get_hardsect_size(s->cow->dev);
 	if (chunk_size % (blocksize / SECTOR_SIZE)) {
 		*context = "Chunk size is not a multiple of device blocksize";
@@ -475,12 +462,6 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 		goto bad_putdev;
 	}
 
-	if (extent_size % (blocksize / SECTOR_SIZE)) {
-		*context = "Extent size is not a multiple of device blocksize";
-		r = -EINVAL;
-		goto bad_putdev;
-	}
-
 	/* Check the sizes are small enough to fit in one kiovec */
 	if (chunk_size > KIO_MAX_SECTORS) {
 		*context = "Chunk size is too big";
@@ -488,12 +469,6 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 		goto bad_putdev;
 	}
 
-	if (extent_size > KIO_MAX_SECTORS) {
-		*context = "Extent size is too big";
-		r = -EINVAL;
-		goto bad_putdev;
-	}
-
 	/* Check chunk_size is a power of 2 */
 	if (chunk_size & (chunk_size - 1)) {
 		*context = "Chunk size is not a power of 2";
@@ -501,7 +476,6 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 		goto bad_putdev;
 	}
 
-	init_waitqueue_head(&s->waitq);
 	s->chunk_size = chunk_size;
 	s->chunk_mask = chunk_size - 1;
 	s->type = *persistent;
@@ -512,6 +486,7 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 
 	s->valid = 1;
 	s->last_percent = 0;
+	s->table = t;
 	init_rwsem(&s->lock);
 
 	/* Allocate hash table for COW data */
@@ -560,7 +535,7 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 	return 0;
 
       bad_free2:
-	store_fn(s, destroy);
+	s->store.destroy(&s->store);
 
       bad_free1:
 	exit_exception_table(&s->pending, pending_cache);
@@ -581,7 +556,7 @@ static void snapshot_dtr(struct dm_table *t, void *context)
 {
 	struct dm_snapshot *s = (struct dm_snapshot *) context;
 
-	wake_up_interruptible(&s->waitq);
+	dm_table_event(s->table);
 
 	unregister_snapshot(s);
 
@@ -589,7 +564,7 @@ static void snapshot_dtr(struct dm_table *t, void *context)
 	exit_exception_table(&s->complete, exception_cache);
 
 	/* Deallocate memory used */
-	store_fn(s, destroy);
+	s->store.destroy(&s->store);
 
 	dm_table_put_device(t, s->origin);
 	dm_table_put_device(t, s->cow);
@@ -651,7 +626,7 @@ static void pending_complete(struct pending_exception *pe, int success)
 		if (!e) {
 			printk("Unable to allocate exception.");
 			down_write(&s->lock);
-			store_fn(s, drop_snapshot);
+			s->store.drop_snapshot(&s->store);
 			s->valid = 0;
 			up_write(&s->lock);
 			return;
@@ -678,7 +653,7 @@ static void pending_complete(struct pending_exception *pe, int success)
 			int pc = s->store.percent_full(&s->store);
 
 			if (pc >= s->last_percent + WAKE_UP_PERCENT) {
-				wake_up_interruptible(&s->waitq);
+				dm_table_event(s->table);
 				s->last_percent = pc - pc % WAKE_UP_PERCENT;
 			}
 		}
@@ -688,14 +663,14 @@ static void pending_complete(struct pending_exception *pe, int success)
 		DMERR("Error reading/writing snapshot");
 
 		down_write(&s->lock);
-		store_fn(s, drop_snapshot);
+		s->store.drop_snapshot(&s->store);
 		s->valid = 0;
 		remove_exception(&pe->e);
 		up_write(&s->lock);
 
 		error_buffers(pe->snapshot_bhs);
 
-		wake_up_interruptible(&s->waitq);
+		dm_table_event(s->table);
 		DMDEBUG("Exception failed.");
 	}
 
@@ -788,7 +763,7 @@ static struct pending_exception *find_pending_exception(struct dm_snapshot *s,
 		pe->snap = s;
 		pe->started = 0;
 
-		if (store_int_fn(s, prepare_exception, &pe->e)) {
+		if (s->store.prepare_exception(&s->store, &pe->e)) {
 			free_pending_exception(pe);
 			s->valid = 0;
 			return NULL;
@@ -840,7 +815,7 @@ static int snapshot_map(struct buffer_head *bh, int rw, void *context)
 			pe = find_pending_exception(s, bh);
 
 			if (!pe) {
-				store_fn(s, drop_snapshot);
+				s->store.drop_snapshot(&s->store);
 				s->valid = 0;
 			}
 
@@ -923,7 +898,7 @@ static int __origin_write(struct list_head *snapshots, struct buffer_head *bh)
 		if (!e) {
 			pe = find_pending_exception(snap, bh);
 			if (!pe) {
-				store_fn(snap, drop_snapshot);
+				snap->store.drop_snapshot(&snap->store);
 				snap->valid = 0;
 
 			} else {
@@ -995,18 +970,6 @@ static int snapshot_status(status_type_t type, char *result,
 	return 0;
 }
 
-static int snapshot_wait(void *context, wait_queue_t *wq, int add)
-{
-	struct dm_snapshot *snap = (struct dm_snapshot *) context;
-
-	if (add)
-		add_wait_queue(&snap->waitq, wq);
-	else
-		remove_wait_queue(&snap->waitq, wq);
-
-	return 0;
-}
-
 /*
  * Called on a write from the origin driver.
  */
@@ -1046,7 +1009,7 @@ static int origin_ctr(struct dm_table *t, offset_t b, offset_t l,
 		return -EINVAL;
 	}
 
-	r = dm_table_get_device(t, argv[0], 0, l, &dev);
+	r = dm_table_get_device(t, argv[0], 0, l, t->mode, &dev);
 	if (r) {
 		*context = "Cannot get target device";
 		return r;
@@ -1097,7 +1060,6 @@ static struct target_type origin_target = {
 	dtr:	origin_dtr,
 	map:	origin_map,
 	status:	origin_status,
-	wait:	NULL,
 	err:	NULL
 };
 
@@ -1108,7 +1070,6 @@ static struct target_type snapshot_target = {
 	dtr:	snapshot_dtr,
 	map:	snapshot_map,
 	status:	snapshot_status,
-	wait:	snapshot_wait,
 	err:	NULL
 };
 
