@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2001 Sistina Software
+ * Copyright (C) 2001 Sistina Software (UK) Limited.
  *
  * This file is released under the GPL.
  */
@@ -7,34 +7,38 @@
 #include "dm.h"
 
 #include <linux/blk.h>
-#include <linux/blkdev.h>
 #include <linux/blkpg.h>
-#include <linux/kmod.h>
 
 /* we only need this for the lv_bmap struct definition, not happy */
 #include <linux/lvm.h>
 
-#define MAX_DEVICES 64
 #define DEFAULT_READ_AHEAD 64
-#define DEVICE_NAME "device-mapper"
 
-static const char *_name = DEVICE_NAME;
-static int _version[3] = { 0, 1, 0 };
+static const char *_name = DM_NAME;
+
 static int major = 0;
+static int _major = 0;
 
 struct io_hook {
 	struct mapped_device *md;
 	struct target *target;
 	int rw;
 
-	void (*end_io)(struct buffer_head *bh, int uptodate);
+	void (*end_io) (struct buffer_head * bh, int uptodate);
 	void *context;
 };
 
 static kmem_cache_t *_io_hook_cache;
 
-static struct rw_semaphore _dev_lock;
 static struct mapped_device *_devs[MAX_DEVICES];
+static struct rw_semaphore _dev_locks[MAX_DEVICES];
+
+/*
+ * This lock is only held by dm_create and dm_set_name to avoid
+ * race conditions where someone else may create a device with
+ * the same name.
+ */
+static spinlock_t _create_lock = SPIN_LOCK_UNLOCKED;
 
 /* block device arrays */
 static int _block_size[MAX_DEVICES];
@@ -43,127 +47,293 @@ static int _hardsect_size[MAX_DEVICES];
 
 static devfs_handle_t _dev_dir;
 
-static int request(request_queue_t *q, int rw, struct buffer_head *bh);
+static int request(request_queue_t * q, int rw, struct buffer_head *bh);
 static int dm_user_bmap(struct inode *inode, struct lv_bmap *lvb);
 
 /*
- * setup and teardown the driver
+ * Protect the mapped_devices referenced from _dev[]
  */
-static int __init dm_init(void)
+struct mapped_device *dm_get_r(int minor)
 {
-	int ret = -ENOMEM;
+	struct mapped_device *md;
 
-	init_rwsem(&_dev_lock);
+	if (minor >= MAX_DEVICES)
+		return NULL;
 
-	_io_hook_cache = kmem_cache_create("dm io hooks",
-					   sizeof(struct io_hook),
-					   0, 0, NULL, NULL);
+	down_read(_dev_locks + minor);
+	md = _devs[minor];
+	if (!md)
+		up_read(_dev_locks + minor);
+	return md;
+}
 
-	if (!_io_hook_cache)
-		goto err;
+struct mapped_device *dm_get_w(int minor)
+{
+	struct mapped_device *md;
 
-	ret = dm_target_init();
-	if (ret < 0)
-		goto err_cache_free;
+	if (minor >= MAX_DEVICES)
+		return NULL;
 
-	ret = dm_interface_init();
-	if (ret < 0)
-		goto err_cache_free;
+	down_write(_dev_locks + minor);
+	md = _devs[minor];
+	if (!md)
+		up_write(_dev_locks + minor);
+	return md;
+}
 
-	ret = devfs_register_blkdev(major, _name, &dm_blk_dops);
-	if (ret < 0)
-		goto err_blkdev;
+static int namecmp(struct mapped_device *md, const char *name, int nametype)
+{
+	switch (nametype) {
+	case DM_LOOKUP_BY_NAME:
+		return strcmp(md->name, name);
+		break;
 
-	if (major == 0)
-		major = ret;
+	case DM_LOOKUP_BY_UUID:
+		if (!md->uuid)
+			return -1;	/* never equal */
+
+		return strcmp(md->uuid, name);
+		break;
+
+	default:
+		DMWARN("Unknown comparison type in namecmp: %d", nametype);
+		BUG();
+	}
+
+	return -1;
+}
+
+/*
+ * The interface (eg, ioctl) will probably access the devices
+ * through these slow 'by name' locks, this needs improving at
+ * some point if people start playing with *large* numbers of dm
+ * devices.
+ */
+struct mapped_device *dm_get_name_r(const char *name, int nametype)
+{
+	int i;
+	struct mapped_device *md;
+
+	for (i = 0; i < MAX_DEVICES; i++) {
+		md = dm_get_r(i);
+		if (md) {
+			if (!namecmp(md, name, nametype))
+				return md;
+
+			dm_put_r(md);
+		}
+	}
+
+	return NULL;
+}
+
+struct mapped_device *dm_get_name_w(const char *name, int nametype)
+{
+	int i;
+	struct mapped_device *md;
+
+	/*
+	 * To avoid getting write locks on all the devices we try
+	 * and promote a read lock to a write lock, this can
+	 * fail, in which case we just start again.
+	 */
+
+      restart:
+	for (i = 0; i < MAX_DEVICES; i++) {
+		md = dm_get_r(i);
+		if (!md)
+			continue;
+
+		if (namecmp(md, name, nametype)) {
+			dm_put_r(md);
+			continue;
+		}
+
+		/* found it */
+		dm_put_r(md);
+
+		md = dm_get_w(i);
+		if (!md)
+			goto restart;
+
+		if (namecmp(md, name, nametype)) {
+			dm_put_w(md);
+			goto restart;
+		}
+
+		return md;
+	}
+
+	return NULL;
+}
+
+void dm_put_r(struct mapped_device *md)
+{
+	int minor = MINOR(md->dev);
+
+	if (minor >= MAX_DEVICES)
+		return;
+
+	up_read(_dev_locks + minor);
+}
+
+void dm_put_w(struct mapped_device *md)
+{
+	int minor = MINOR(md->dev);
+
+	if (minor >= MAX_DEVICES)
+		return;
+
+	up_write(_dev_locks + minor);
+}
+
+/*
+ * Setup and tear down the driver
+ */
+static __init void init_locks(void)
+{
+	int i;
+
+	for (i = 0; i < MAX_DEVICES; i++)
+		init_rwsem(_dev_locks + i);
+}
+
+static __init int local_init(void)
+{
+	int r;
+
+	init_locks();
+
+	/* allocate a slab for the io-hooks */
+	if (!_io_hook_cache &&
+	    !(_io_hook_cache = kmem_cache_create("dm io hooks",
+						 sizeof(struct io_hook),
+						 0, 0, NULL, NULL)))
+		return -ENOMEM;
+
+	_major = major;
+	r = devfs_register_blkdev(_major, _name, &dm_blk_dops);
+	if (r < 0) {
+		DMERR("register_blkdev failed");
+		kmem_cache_destroy(_io_hook_cache);
+		return r;
+	}
+
+	if (!_major)
+		_major = r;
 
 	/* set up the arrays */
-	read_ahead[major] = DEFAULT_READ_AHEAD;
-	blk_size[major] = _block_size;
-	blksize_size[major] = _blksize_size;
-	hardsect_size[major] = _hardsect_size;
+	read_ahead[_major] = DEFAULT_READ_AHEAD;
+	blk_size[_major] = _block_size;
+	blksize_size[_major] = _blksize_size;
+	hardsect_size[_major] = _hardsect_size;
 
-	blk_queue_make_request(BLK_DEFAULT_QUEUE(major), request);
+	blk_queue_make_request(BLK_DEFAULT_QUEUE(_major), request);
 
 	_dev_dir = devfs_mk_dir(0, DM_DIR, NULL);
 
-	printk(KERN_INFO "%s %d.%d.%d initialised\n", _name,
-	       _version[0], _version[1], _version[2]);
+	return 0;
+}
+
+static void local_exit(void)
+{
+	if (kmem_cache_destroy(_io_hook_cache))
+		DMWARN("io_hooks still allocated during unregistration");
+	_io_hook_cache = NULL;
+
+	if (devfs_unregister_blkdev(_major, _name) < 0)
+		DMERR("devfs_unregister_blkdev failed");
+
+	read_ahead[_major] = 0;
+	blk_size[_major] = NULL;
+	blksize_size[_major] = NULL;
+	hardsect_size[_major] = NULL;
+	_major = 0;
+
+	DMINFO("cleaned up");
+}
+
+/*
+ * We have a lot of init/exit functions, so it seems easier to
+ * store them in an array.  The disposable macro 'xx'
+ * expands a prefix into a pair of function names.
+ */
+static struct {
+	int (*init)(void);
+	void (*exit)(void);
+
+} _inits[] = {
+#define xx(n) {n ## _init, n ## _exit},
+	xx(local)
+	xx(dm_target)
+	xx(dm_linear)
+	xx(dm_stripe)
+	xx(dm_snapshot)
+/*	xx(dm_mirror) */
+	xx(dm_interface)
+#undef xx
+};
+
+static int __init dm_init(void)
+{
+	const int count = sizeof(_inits) / sizeof(*_inits);
+
+	int r, i;
+
+	for (i = 0; i < count; i++) {
+		r = _inits[i].init();
+		if (r)
+			goto bad;
+	}
+
 	return 0;
 
-      err_blkdev:
-	printk(KERN_ERR "%s -- register_blkdev failed\n", _name);
-	dm_interface_exit();
-      err_cache_free:
-	kmem_cache_destroy(_io_hook_cache);
-      err:
-	return ret;
+      bad:
+	while (i--)
+		_inits[i].exit();
+
+	return r;
 }
 
 static void __exit dm_exit(void)
 {
-	dm_interface_exit();
+	int i = sizeof(_inits) / sizeof(*_inits);
 
-	if (kmem_cache_destroy(_io_hook_cache))
-		WARN("it looks like there are still some io_hooks allocated");
-
-	_io_hook_cache = NULL;
-
-	if (devfs_unregister_blkdev(major, _name) < 0)
-		printk(KERN_ERR "%s -- unregister_blkdev failed\n", _name);
-
-	read_ahead[major] = 0;
-	blk_size[major] = NULL;
-	blksize_size[major] = NULL;
-	hardsect_size[major] = NULL;
-
-	printk(KERN_INFO "%s %d.%d.%d cleaned up\n", _name,
-	       _version[0], _version[1], _version[2]);
+	dm_destroy_all();
+	while (i--)
+		_inits[i].exit();
 }
 
 /*
- * block device functions
+ * Block device functions
  */
 static int dm_blk_open(struct inode *inode, struct file *file)
 {
-	int minor = MINOR(inode->i_rdev);
 	struct mapped_device *md;
 
-	if (minor >= MAX_DEVICES)
+	md = dm_get_w(MINOR(inode->i_rdev));
+	if (!md)
 		return -ENXIO;
 
-	down_write(&_dev_lock);
-	md = _devs[minor];
-
-	if (!md) {
-		up_write(&_dev_lock);
-		return -ENXIO;
-	}
-
 	md->use_count++;
-	up_write(&_dev_lock);
+	dm_put_w(md);
 
 	return 0;
 }
 
 static int dm_blk_close(struct inode *inode, struct file *file)
 {
-	int minor = MINOR(inode->i_rdev);
 	struct mapped_device *md;
 
-	if (minor >= MAX_DEVICES)
+	md = dm_get_w(MINOR(inode->i_rdev));
+	if (!md)
 		return -ENXIO;
 
-	down_write(&_dev_lock);
-	md = _devs[minor];
-	if (!md || md->use_count < 1) {
-		WARN("reference count in mapped_device incorrect");
-		up_write(&_dev_lock);
-		return -ENXIO;
-	}
+	if (md->use_count < 1)
+		DMWARN("incorrect reference count found in mapped_device");
 
 	md->use_count--;
-	up_write(&_dev_lock);
+	dm_put_w(md);
 
 	return 0;
 }
@@ -181,41 +351,42 @@ static int dm_blk_ioctl(struct inode *inode, struct file *file,
 		return -ENXIO;
 
 	switch (command) {
-	case BLKSSZGET:
-	case BLKBSZGET:
-	case BLKROGET:
 	case BLKROSET:
+	case BLKROGET:
 	case BLKRASET:
 	case BLKRAGET:
 	case BLKFLSBUF:
-#if 0   /* Future stacking block device */
-	case BLKELVSET:
+	case BLKSSZGET:
+		//case BLKRRPART: /* Re-read partition tables */
+		//case BLKPG:
 	case BLKELVGET:
-#endif
+	case BLKELVSET:
+	case BLKBSZGET:
+	case BLKBSZSET:
 		return blk_ioctl(inode->i_rdev, command, a);
 		break;
 
 	case BLKGETSIZE:
 		size = VOLUME_SIZE(minor);
-		if (copy_to_user((void *)a, &size, sizeof(long)))
+		if (copy_to_user((void *) a, &size, sizeof(long)))
 			return -EFAULT;
 		break;
 
 	case BLKGETSIZE64:
 		size = VOLUME_SIZE(minor);
-		if (put_user((u64)size, (u64 *)a))
+		if (put_user((u64) ((u64) size) << 9, (u64 *) a))
 			return -EFAULT;
 		break;
 
 	case BLKRRPART:
-		return -EINVAL;
+		return -ENOTTY;
 
 	case LV_BMAP:
 		return dm_user_bmap(inode, (struct lv_bmap *) a);
 
 	default:
-		WARN("unknown block ioctl %d", command);
-		return -EINVAL;
+		DMWARN("unknown block ioctl 0x%x", command);
+		return -ENOTTY;
 	}
 
 	return 0;
@@ -232,7 +403,7 @@ static inline void free_io_hook(struct io_hook *ih)
 }
 
 /*
- * FIXME: need to decide if deferred_io's need
+ * FIXME: We need to decide if deferred_io's need
  * their own slab, I say no for now since they are
  * only used when the device is suspended.
  */
@@ -247,12 +418,12 @@ static inline void free_deferred(struct deferred_io *di)
 }
 
 /*
- * call a targets optional error function if
- * an io failed.
+ * Call a target's optional error function if an I/O failed.
  */
 static inline int call_err_fn(struct io_hook *ih, struct buffer_head *bh)
 {
 	dm_err_fn err = ih->target->type->err;
+
 	if (err)
 		return err(bh, ih->rw, ih->target->private);
 
@@ -260,9 +431,8 @@ static inline int call_err_fn(struct io_hook *ih, struct buffer_head *bh)
 }
 
 /*
- * bh->b_end_io routine that decrements the
- * pending count and then calls the original
- * bh->b_end_io fn.
+ * bh->b_end_io routine that decrements the pending count
+ * and then calls the original bh->b_end_io fn.
  */
 static void dec_pending(struct buffer_head *bh, int uptodate)
 {
@@ -283,32 +453,40 @@ static void dec_pending(struct buffer_head *bh, int uptodate)
 }
 
 /*
- * add the bh to the list of deferred io.
+ * Add the bh to the list of deferred io.
  */
-static int queue_io(struct mapped_device *md, struct buffer_head *bh, int rw)
+static int queue_io(struct buffer_head *bh, int rw)
 {
 	struct deferred_io *di = alloc_deferred();
+	struct mapped_device *md;
 
 	if (!di)
 		return -ENOMEM;
 
-	down_write(&_dev_lock);
+	md = dm_get_w(MINOR(bh->b_rdev));
+	if (!md) {
+		free_deferred(di);
+		return -ENXIO;
+	}
+
 	if (!md->suspended) {
-		up_write(&_dev_lock);
-		return 0;
+		dm_put_w(md);
+		free_deferred(di);
+		return 1;
 	}
 
 	di->bh = bh;
 	di->rw = rw;
 	di->next = md->deferred;
 	md->deferred = di;
-	up_write(&_dev_lock);
 
-	return 1;
+	dm_put_w(md);
+
+	return 0;		/* deferred successfully */
 }
 
 /*
- * do the bh mapping for a given leaf
+ * Do the bh mapping for a given leaf
  */
 static inline int __map_buffer(struct mapped_device *md,
 			       struct buffer_head *bh, int rw, int leaf)
@@ -325,7 +503,7 @@ static inline int __map_buffer(struct mapped_device *md,
 	ih = alloc_io_hook();
 
 	if (!ih)
-		return 0;
+		return -1;
 
 	ih->md = md;
 	ih->rw = rw;
@@ -347,14 +525,14 @@ static inline int __map_buffer(struct mapped_device *md,
 
 	else if (r < 0) {
 		free_io_hook(ih);
-		return 0;
+		return -1;
 	}
 
-	return 1;
+	return r;
 }
 
 /*
- * search the btree for the correct target.
+ * Search the btree for the correct target.
  */
 static inline int __find_node(struct dm_table *t, struct buffer_head *bh)
 {
@@ -373,55 +551,68 @@ static inline int __find_node(struct dm_table *t, struct buffer_head *bh)
 	return (KEYS_PER_NODE * n) + k;
 }
 
-static int request(request_queue_t *q, int rw, struct buffer_head *bh)
+static int request(request_queue_t * q, int rw, struct buffer_head *bh)
 {
 	struct mapped_device *md;
 	int r, minor = MINOR(bh->b_rdev);
+	unsigned int block_size = _blksize_size[minor];
 
-	if (minor >= MAX_DEVICES)
-		goto bad_no_lock;
+	md = dm_get_r(minor);
+	if (!md) {
+		buffer_IO_error(bh);
+		return 0;
+	}
 
-	down_read(&_dev_lock);
-	md = _devs[minor];
+	/*
+	 * Sanity checks.
+	 */
+	if (bh->b_size > block_size)
+		DMERR("request is larger than block size "
+		      "b_size (%d), block size (%d)",
+		      bh->b_size, block_size);
 
-	if (!md)
-		goto bad;
+	if (bh->b_rsector & ((bh->b_size >> 9) - 1))
+		DMERR("misaligned block requested logical "
+		      "sector (%lu), b_size (%d)",
+		      bh->b_rsector, bh->b_size);
 
 	/*
 	 * If we're suspended we have to queue
 	 * this io for later.
 	 */
 	while (md->suspended) {
-		up_read(&_dev_lock);
+		dm_put_r(md);
 
 		if (rw == READA)
 			goto bad_no_lock;
 
-		r = queue_io(md, bh, rw);
+		r = queue_io(bh, rw);
 
 		if (r < 0)
 			goto bad_no_lock;
 
-		else if (r > 0)
+		else if (r == 0)
 			return 0;	/* deferred successfully */
 
 		/*
-		 * We're in a while loop, because
-		 * someone could suspend before we
-		 * get to the following read
-		 * lock
+		 * We're in a while loop, because someone could suspend
+		 * before we get to the following read lock.
 		 */
-		down_read(&_dev_lock);
+		md = dm_get_r(minor);
+		if (!md) {
+			buffer_IO_error(bh);
+			return 0;
+		}
 	}
 
-	if (!__map_buffer(md, bh, rw, __find_node(md->map, bh)))
+	if ((r = __map_buffer(md, bh, rw, __find_node(md->map, bh))) < 0)
 		goto bad;
 
-	up_read(&_dev_lock);
-	return 1;
+	dm_put_r(md);
+	return r;
 
       bad:
-	up_read(&_dev_lock);
+	dm_put_r(md);
 
       bad_no_lock:
 	buffer_IO_error(bh);
@@ -438,7 +629,7 @@ static int check_dev_size(int minor, unsigned long block)
 }
 
 /*
- * creates a dummy buffer head and maps it (for lilo).
+ * Creates a dummy buffer head and maps it (for lilo).
  */
 static int do_bmap(kdev_t dev, unsigned long block,
 		   kdev_t * r_dev, unsigned long *r_block)
@@ -448,15 +639,18 @@ static int do_bmap(kdev_t dev, unsigned long block,
 	int minor = MINOR(dev), r;
 	struct target *t;
 
-	down_read(&_dev_lock);
-	if ((minor >= MAX_DEVICES) || !(md = _devs[minor]) || md->suspended) {
-		r = -ENXIO;
-		goto out;
+	md = dm_get_r(minor);
+	if (!md)
+		return -ENXIO;
+
+	if (md->suspended) {
+		dm_put_r(md);
+		return -EPERM;
 	}
 
 	if (!check_dev_size(minor, block)) {
-		r = -EINVAL;
-		goto out;
+		dm_put_r(md);
+		return -EINVAL;
 	}
 
 	/* setup dummy bh */
@@ -475,14 +669,12 @@ static int do_bmap(kdev_t dev, unsigned long block,
 	*r_dev = bh.b_rdev;
 	*r_block = bh.b_rsector / (bh.b_size >> 9);
 
-      out:
-	up_read(&_dev_lock);
+	dm_put_r(md);
 	return r;
 }
 
 /*
- * marshals arguments and results between user and
- * kernel space.
+ * Marshals arguments and results between user and kernel space.
  */
 static int dm_user_bmap(struct inode *inode, struct lv_bmap *lvb)
 {
@@ -497,83 +689,100 @@ static int dm_user_bmap(struct inode *inode, struct lv_bmap *lvb)
 		return r;
 
 	if (put_user(kdev_t_to_nr(r_dev), &lvb->lv_dev) ||
-	    put_user(r_block, &lvb->lv_block)) 
+	    put_user(r_block, &lvb->lv_block))
 		return -EFAULT;
 
 	return 0;
 }
 
 /*
- * see if the device with a specific minor # is
- * free.
+ * See if the device with a specific minor # is free.  The write
+ * lock is held when it returns successfully.
  */
-static inline int __specific_dev(int minor)
+static inline int specific_dev(int minor, struct mapped_device *md)
 {
-	if (minor > MAX_DEVICES) {
-		WARN("request for a mapped_device > than MAX_DEVICES");
-		return 0;
+	if (minor >= MAX_DEVICES) {
+		DMWARN("request for a mapped_device beyond MAX_DEVICES (%d)",
+		       MAX_DEVICES);
+		return -1;
 	}
 
-	if (!_devs[minor])
-		return minor;
+	down_write(_dev_locks + minor);
+	if (_devs[minor]) {
+		/* in use */
+		up_write(_dev_locks + minor);
+		return -1;
+	}
 
-	return -1;
+	return minor;
 }
 
 /*
- * find the first free device.
+ * Find the first free device.  Again the write lock is held on
+ * success.
  */
-static inline int __any_old_dev(void)
+static int any_old_dev(struct mapped_device *md)
 {
 	int i;
 
 	for (i = 0; i < MAX_DEVICES; i++)
-		if (!_devs[i])
+		if (specific_dev(i, md) != -1)
 			return i;
 
 	return -1;
 }
 
 /*
- * allocate and initialise a blank device.
+ * Allocate and initialise a blank device.
+ * Caller must ensure uuid is null-terminated.
+ * Device is returned with a write lock held.
  */
-static struct mapped_device *alloc_dev(int minor)
+static struct mapped_device *alloc_dev(const char *name, const char *uuid,
+				       int minor)
 {
 	struct mapped_device *md = kmalloc(sizeof(*md), GFP_KERNEL);
+	int len;
 
-	if (!md)
-		return 0;
+	if (!md) {
+		DMWARN("unable to allocate device, out of memory.");
+		return NULL;
+	}
 
 	memset(md, 0, sizeof(*md));
 
-	down_write(&_dev_lock);
-	minor = (minor < 0) ? __any_old_dev() : __specific_dev(minor);
-
+	/*
+	 * This grabs the write lock if it succeeds.
+	 */
+	minor = (minor < 0) ? any_old_dev(md) : specific_dev(minor, md);
 	if (minor < 0) {
-		WARN("no free devices available");
-		up_write(&_dev_lock);
 		kfree(md);
-		return 0;
+		return NULL;
 	}
 
-	md->dev = MKDEV(major, minor);
-	md->name[0] = '\0';
+	md->dev = MKDEV(_major, minor);
 	md->suspended = 0;
 
+	strncpy(md->name, name, sizeof(md->name) - 1);
+	md->name[sizeof(md->name) - 1] = '\0';
+
+	/*
+	 * Copy in the uuid.
+	 */
+	if (uuid && *uuid) {
+		len = strlen(uuid) + 1;
+		if (!(md->uuid = kmalloc(len, GFP_KERNEL))) {
+			DMWARN("unable to allocate uuid - out of memory.");
+			kfree(md);
+			return NULL;
+		}
+		strcpy(md->uuid, uuid);
+	}
+
 	init_waitqueue_head(&md->wait);
-
-	_devs[minor] = md;
-	up_write(&_dev_lock);
-
 	return md;
 }
 
-static void free_dev(struct mapped_device *md)
-{
-	kfree(md);
-}
-
-static int register_device(struct mapped_device *md)
+static int __register_device(struct mapped_device *md)
 {
 	md->devfs_entry =
 	    devfs_register(_dev_dir, md->name, DEVFS_FL_CURRENT_OWNER,
@@ -584,16 +793,15 @@ static int register_device(struct mapped_device *md)
 	return 0;
 }
 
-static int unregister_device(struct mapped_device *md)
+static int __unregister_device(struct mapped_device *md)
 {
 	devfs_unregister(md->devfs_entry);
 	return 0;
 }
 
 /*
- * the hardsect size for a mapped device is the
- * smallest hard sect size from the devices it
- * maps onto.
+ * The hardsect size for a mapped device is the smallest hardsect size
+ * from the devices it maps onto.
  */
 static int __find_hardsect_size(struct list_head *devices)
 {
@@ -606,6 +814,14 @@ static int __find_hardsect_size(struct list_head *devices)
 		if (size < result)
 			result = size;
 	}
+
+	/*
+	 * I think it's safe to assume that no block devices have
+	 * a hard sector size this large.
+	 */
+	if (result == INT_MAX)
+		result = 512;
+
 	return result;
 }
 
@@ -647,56 +863,65 @@ static void __unbind(struct mapped_device *md)
 	_hardsect_size[minor] = 0;
 }
 
-static struct mapped_device *__get_by_name(const char *name)
-{
-	int i;
-
-	for (i = 0; i < MAX_DEVICES; i++)
-		if (_devs[i] && !strcmp(_devs[i]->name, name))
-			return _devs[i];
-
-	return NULL;
-}
-
 static int check_name(const char *name)
 {
-	if (strchr(name, '/')) {
-		WARN("invalid device name");
-		return 0;
+	struct mapped_device *md;
+
+	if (strchr(name, '/') || strlen(name) > DM_NAME_LEN) {
+		DMWARN("invalid device name");
+		return -1;
 	}
 
-	if (__get_by_name(name)) {
-		WARN("device name already in use");
-		return 0;
+	md = dm_get_name_r(name, DM_LOOKUP_BY_NAME);
+	if (md) {
+		dm_put_r(md);
+		DMWARN("device name already in use");
+		return -1;
 	}
 
-	return 1;
+	return 0;
+}
+
+static int check_uuid(const char *uuid)
+{
+	struct mapped_device *md;
+
+	if (uuid) {
+		md = dm_get_name_r(uuid, DM_LOOKUP_BY_UUID);
+		if (md) {
+			dm_put_r(md);
+			DMWARN("device uuid already in use");
+			return -1;
+		}
+	}
+
+	return 0;
 }
 
 /*
- * constructor for a new device
+ * Constructor for a new device.
  */
-struct mapped_device *dm_create(const char *name, int minor,
-				struct dm_table *table)
+int dm_create(const char *name, const char *uuid, int minor, int ro,
+	      struct dm_table *table)
 {
-	int r = -EINVAL;
+	int r;
 	struct mapped_device *md;
 
-	if (minor >= MAX_DEVICES)
-		return ERR_PTR(-ENXIO);
+	spin_lock(&_create_lock);
+	if (check_name(name) || check_uuid(uuid)) {
+		spin_unlock(&_create_lock);
+		return -EINVAL;
+	}
 
-	if (!(md = alloc_dev(minor)))
-		return ERR_PTR(-ENXIO);
-
-	down_write(&_dev_lock);
-
-	if (!check_name(name))
-		goto err;
-
-	strcpy(md->name, name);
+	md = alloc_dev(name, uuid, minor);
+	if (!md) {
+		spin_unlock(&_create_lock);
+		return -ENXIO;
+	}
+	minor = MINOR(md->dev);
 	_devs[minor] = md;
 
-	r = register_device(md);
+	r = __register_device(md);
 	if (r)
 		goto err;
 
@@ -704,58 +929,127 @@ struct mapped_device *dm_create(const char *name, int minor,
 	if (r)
 		goto err;
 
-	up_write(&_dev_lock);
+	dm_set_ro(md, ro);
 
-	return md;
+	spin_unlock(&_create_lock);
+	dm_put_w(md);
+	return 0;
 
       err:
-	up_write(&_dev_lock);
-	free_dev(md);
-	return ERR_PTR(r);
+	_devs[minor] = NULL;
+	if (md->uuid)
+		kfree(md->uuid);
+
+	dm_put_w(md);
+	kfree(md);
+	spin_unlock(&_create_lock);
+	return r;
 }
 
 /*
- * Destructor for the device.  You cannot destroy
- * a suspended device.
+ * Renames the device.  No lock held.
+ */
+int dm_set_name(const char *name, int nametype, const char *newname)
+{
+	int r;
+	struct mapped_device *md;
+
+	spin_lock(&_create_lock);
+	if (check_name(newname) < 0) {
+		spin_unlock(&_create_lock);
+		return -EINVAL;
+	}
+
+	md = dm_get_name_w(name, nametype);
+	if (!md) {
+		spin_unlock(&_create_lock);
+		return -ENXIO;
+	}
+
+	r = __unregister_device(md);
+	if (r)
+		goto out;
+
+	strcpy(md->name, newname);
+	r = __register_device(md);
+
+      out:
+	dm_put_w(md);
+	spin_unlock(&_create_lock);
+	return r;
+}
+
+/*
+ * Destructor for the device.  You cannot destroy an open
+ * device.  Write lock must be held before calling.
+ * Caller must dm_put_w(md) then kfree(md) if call was successful.
  */
 int dm_destroy(struct mapped_device *md)
 {
 	int minor, r;
 
-	down_read(&_dev_lock);
-	if (md->suspended || md->use_count) {
-		up_read(&_dev_lock);
+	if (md->use_count)
 		return -EPERM;
-	}
 
-	fsync_dev(md->dev);
-	up_read(&_dev_lock);
-
-	down_write(&_dev_lock);
-	if (md->use_count) {
-		up_write(&_dev_lock);
-		return -EPERM;
-	}
-
-	if ((r = unregister_device(md))) {
-		up_write(&_dev_lock);
+	r = __unregister_device(md);
+	if (r)
 		return r;
-	}
 
 	minor = MINOR(md->dev);
-	_devs[minor] = 0;
+	_devs[minor] = NULL;
 	__unbind(md);
 
-	up_write(&_dev_lock);
-
-	free_dev(md);
+	if (md->uuid)
+		kfree(md->uuid);
 
 	return 0;
 }
 
 /*
- * requeue the deferred buffer_heads by calling
- * generic_make_request.
+ * Destroy all devices - except open ones
+ */
+void dm_destroy_all(void)
+{
+	int i, some_destroyed, r;
+	struct mapped_device *md;
+
+	do {
+		some_destroyed = 0;
+		for (i = 0; i < MAX_DEVICES; i++) {
+			md = dm_get_w(i);
+			if (!md)
+				continue;
+
+			r = dm_destroy(md);
+			dm_put_w(md);
+
+			if (!r) {
+				kfree(md);
+				some_destroyed = 1;
+			}
+		}
+	} while (some_destroyed);
+}
+
+/*
+ * Sets or clears the read-only flag for the device.  Write lock
+ * must be held.
+ */
+void dm_set_ro(struct mapped_device *md, int ro)
+{
+	md->read_only = ro;
+	set_device_ro(md->dev, ro);
+}
+
+/*
+ * A target is notifying us of some event
+ */
+void dm_notify(void *target)
+{
+}
+
+/*
+ * Requeue the deferred buffer_heads by calling generic_make_request.
  */
 static void flush_deferred_io(struct deferred_io *c)
 {
@@ -770,108 +1064,95 @@ static void flush_deferred_io(struct deferred_io *c)
 }
 
 /*
- * Swap in a new table (destroying old one).
+ * Swap in a new table (destroying old one).  Write lock must be
+ * held.
  */
 int dm_swap_table(struct mapped_device *md, struct dm_table *table)
 {
 	int r;
 
-	down_write(&_dev_lock);
-
 	/* device must be suspended */
-	if (!md->suspended) {
-		up_write(&_dev_lock);
+	if (!md->suspended)
 		return -EPERM;
-	}
 
 	__unbind(md);
 
-	if ((r = __bind(md, table))) {
-		up_write(&_dev_lock);
+	r = __bind(md, table);
+	if (r)
 		return r;
-	}
-
-	up_write(&_dev_lock);
 
 	return 0;
 }
 
 /*
- * We need to be able to change a mapping table
- * under a mounted filesystem.  for example we
- * might want to move some data in the background.
- * Before the table can be swapped with
- * dm_bind_table, dm_suspend must be called to
- * flush any in flight buffer_heads and ensure
- * that any further io gets deferred.
+ * We need to be able to change a mapping table under a mounted
+ * filesystem.  for example we might want to move some data in
+ * the background.  Before the table can be swapped with
+ * dm_bind_table, dm_suspend must be called to flush any in
+ * flight buffer_heads and ensure that any further io gets
+ * deferred.  Write lock must be held.
  */
 int dm_suspend(struct mapped_device *md)
 {
+	int minor = MINOR(md->dev);
 	DECLARE_WAITQUEUE(wait, current);
 
-	down_write(&_dev_lock);
-	if (md->suspended) {
-		up_write(&_dev_lock);
+	if (md->suspended)
 		return -EINVAL;
-	}
 
 	md->suspended = 1;
-	up_write(&_dev_lock);
+	dm_put_w(md);
 
 	/* wait for all the pending io to flush */
 	add_wait_queue(&md->wait, &wait);
 	current->state = TASK_UNINTERRUPTIBLE;
 	do {
-		down_write(&_dev_lock);
+		md = dm_get_w(minor);
+		if (!md) {
+			/* Caller expects to free this lock. Yuck. */
+			down_write(_dev_locks + minor);
+			return -ENXIO;
+		}
+
 		if (!atomic_read(&md->pending))
 			break;
 
-		up_write(&_dev_lock);
+		dm_put_w(md);
 		schedule();
 
 	} while (1);
 
 	current->state = TASK_RUNNING;
 	remove_wait_queue(&md->wait, &wait);
-	up_write(&_dev_lock);
 
 	return 0;
 }
 
 int dm_resume(struct mapped_device *md)
 {
+	int minor = MINOR(md->dev);
 	struct deferred_io *def;
 
-	down_write(&_dev_lock);
-	if (!md->suspended) {
-		up_write(&_dev_lock);
+	if (!md->suspended || !md->map->num_targets)
 		return -EINVAL;
-	}
 
 	md->suspended = 0;
 	def = md->deferred;
 	md->deferred = NULL;
-	up_write(&_dev_lock);
 
+	dm_put_w(md);
 	flush_deferred_io(def);
+	run_task_queue(&tq_disk);
+
+	if (!dm_get_w(minor)) {
+		/* FIXME: yuck */
+		down_write(_dev_locks + minor);
+		return -ENXIO;
+	}
 
 	return 0;
 }
 
-/*
- * Search for a device with a particular name.
- */
-struct mapped_device *dm_get(const char *name)
-{
-	struct mapped_device *md;
-
-	down_read(&_dev_lock);
-	md = __get_by_name(name);
-	up_read(&_dev_lock);
-
-	return md;
-}
-
 struct block_device_operations dm_blk_dops = {
 	open:		dm_blk_open,
 	release:	dm_blk_close,
@@ -887,6 +1168,6 @@ module_exit(dm_exit);
 
 MODULE_PARM(major, "i");
 MODULE_PARM_DESC(major, "The major number of the device mapper");
-MODULE_DESCRIPTION("device-mapper driver");
+MODULE_DESCRIPTION(DM_NAME " driver");
 MODULE_AUTHOR("Joe Thornber <thornber@sistina.com>");
 MODULE_LICENSE("GPL");
